﻿using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;

// <summary>
// This sample is a revision of the NetworkStream Server sample in
// Chapter 02. This server is designed to accept multiple connections
// and demonstrates how to use the NetworkStream class to perform IO between
// 2 sockets. This application performs IO asynchronously by using threads
// where a thread is created for each connection and IO is processed
// synchronously in each thread. The original sample only accepted 1 
// connection and processed IO synchronously. Because this application 
// processes IO synchronously using 1 thread per connection, the application
// is capable of handling multiple connections at the same time.
//  
// To run this sample, simply just run the program without parameters and
// it will listen for a client connection on TCP port 5150. If you want to
// use a different port than 5150 then you can optionally supply a command
// line parameter "/port <port number>" and the listening socket will use
// your port instead. 
// </summary>

namespace ThreadNetworkIO
{
    class Program
    {
        // <summary>
        // The main entry point for the application.
        // </summary>
        [STAThread]
        static void Main(string[] args)
        {
            int Port = 5150;

            // Parse command line arguments if any
            for (int i = 0; i < args.Length; i++)
            {
                try
                {
                    if (String.Compare(args[i], "/port", true) == 0)
                    {
                        // The port on which the server is listening
                        Port = System.Convert.ToInt32(args[++i].ToString());
                    }
                }
                catch (System.IndexOutOfRangeException)
                {
                    Console.WriteLine("Usage: Receiver [/port <port number>]");
                    return;
                }
            }

            Socket ServerSocket = null;
            Socket ListeningSocket = null;

            try
            {
                // Setup a listening Socket to await a connection from a peer socket.
                ListeningSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.IP);
                Console.WriteLine("Socket() is OK...");
                IPEndPoint ListeningEndPoint = new IPEndPoint(IPAddress.Any, Port);
                Console.WriteLine("IPEndPoint() is OK...");
                ListeningSocket.Bind(ListeningEndPoint);
                Console.WriteLine("Bind() is OK...");
                ListeningSocket.Listen(5);
                Console.WriteLine("Listen() is OK...");

                Console.WriteLine("Awaiting TCP connections on IP: "
                    + ListeningEndPoint.Address.ToString()
                    + " Port: "
                    + ListeningEndPoint.Port.ToString()
                    + "...");

                int ConnectionCount = 0;

                while (true)
                {
                    ServerSocket = ListeningSocket.Accept();
                    Console.WriteLine("Accept() is OK...");
                    ConnectionCount++;
                    Console.WriteLine("Connection #" + ConnectionCount.ToString() + " is established and awaiting data...");
                    NetworkStreamHandler NSH = new NetworkStreamHandler(ServerSocket, ConnectionCount);
                    Console.WriteLine("NetworkStreamHandler() is OK...");
                }
            }
            catch (SocketException e)
            {
                Console.WriteLine("Failure to create Sockets: " + e.Message);
            }
            finally
            {
                // Close the listening socket - we do not plan to handle any
                // additional connections.
                ListeningSocket.Close();
            }
        }

        public class NetworkStreamHandler
        {
            // Let's create a network stream to communicate over the connected Socket.
            private NetworkStream ServerNetworkStream = null;
            private int ConnectionNumber = 0;

            public NetworkStreamHandler(Socket ServerSocket, int ConnectionID)
            {
                ConnectionNumber = ConnectionID;
                try
                {
                    // Setup a network stream on the server Socket
                    ServerNetworkStream = new NetworkStream(ServerSocket, true);
                    Console.WriteLine("NetworkStream() is OK...");
                }
                catch (Exception e)
                {
                    Console.WriteLine("Network stream failed with error: " + e.Message);
                    return;
                }

                try
                {
                    ThreadStart NetworkThreadMethod = new ThreadStart(ThreadHandlerMethod);
                    Console.WriteLine("ThreadStart() is OK...");
                    Thread NetworkThread = new Thread(NetworkThreadMethod);
                    Console.WriteLine("Thread() is OK...");
                    NetworkThread.Start();
                    Console.WriteLine("Start() is OK...");
                }
                catch (Exception e)
                {
                    Console.WriteLine("The network thread failed with error: " + e.Message);
                    ServerNetworkStream.Close();
                }
            }

            private void ThreadHandlerMethod()
            {
                try
                {
                    byte[] Buffer = new byte[4096];
                    int BytesRead = 0;

                    do
                    {
                        BytesRead = ServerNetworkStream.Read(Buffer, 0, Buffer.Length);
                        Console.WriteLine();
                        Console.WriteLine("Connection #" + ConnectionNumber.ToString() + " received " + BytesRead.ToString() + " byte(s).");
                    }
                    while (BytesRead > 0);

                }
                catch (ThreadAbortException e)
                {
                    Console.WriteLine("Caught thread abort exception: " + e.Message);
                }
                catch (Exception e)
                {
                    Console.WriteLine("Thread routine failed with error: " + e.Message);
                }
                finally
                {
                    Console.WriteLine("Connection #" + ConnectionNumber.ToString() + " is closing.");
                    ServerNetworkStream.Close();
                }
            }
        }
    }
}
